
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */


#ifndef __MATH_ENUMS_H__
#define __MATH_ENUMS_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "./mp-serializer.h" */
GType math_mp_display_format_get_type (void) G_GNUC_CONST;
#define MP_TYPE_DISPLAY_FORMAT (math_mp_display_format_get_type ())
G_END_DECLS

#endif /* __MATH_ENUMS_H__ */

/* Generated data ends here */

